#include <iostream>  // std::cout
#include <fstream>   // std::ofstream
#include <string>    // std::srting

#include <algorithm> // std::generate
#include <vector>    // std::vector
#include <list>      // std::list
#include <map>       // std::map

// Using extended characters such as the mathematical symbols (\x2200 - \22FF) causes the
// output to halt in the wide file stream.
// *  Is this a bug in my code or the MSVC library?

// tested on: MSVC 7 using the supplied standard library; WinXP; Pentium II

#define EXTENDED_CHARACTER_BUG

#include <boost/io/formatlist.hpp> // boost::io::formatlist

// This is needed for the std::map examples
// NOTE: Is there a better way to do this? io_format?

namespace std
{
   template< typename CharT, class TraitsT, typename T1, typename T2 >
   inline basic_ostream< CharT, TraitsT > & operator<<
   (
      basic_ostream< CharT, TraitsT > & os,
      pair< T1, T2 > & p
   )
   {
      return( os << "( " << p.first << ", " << p.second << " )" );
   }
}

int __cdecl main()
{
   // std::vector

   std::vector< int >                  v( 10 );
   std::generate( v.begin(), v.end(), std::rand );

   // std::list

   std::list< char >                   l;
   l.push_back( 'A' );
   l.push_back( 'B' );
   l.push_back( 'C' );

   // std::map

   std::map< std::string, std::string > m;
   m[ "Abscheid Nehmen"    ] = "Farewell";
   m[ "Engel"              ] = "Angel";
   m[ "Mas Alla"           ] = "Beyond";
   m[ "Wenn Das Liebe Ist" ] = "If This Is Love";

#  if (FORMATLIST__VERSION >= 0x0104) // new examples 04/04/2003:
      std::list< std::wstring >        wl;
      wl.push_back( L"Solange ich Dich fhle," );
      wl.push_back( L"Solange ich Dich spr'," );
      wl.push_back( L"Solange ich in mir whle," );
      wl.push_back( L"Solange," );
      wl.push_back( L"Solang glaub' ich, ich fhr Gutes im Sinn." );
#  endif

   // std::string and arrays

   const char                        * str  = "Jeanette Biedermann";
   std::string                         s    = "Warum bist du nicht da?";

   int                                 i[ 4 ] = { 3, 6, 9, 12 };
   int                                 a[ 3 ][ 3 ] =
   {
      { 1, 0, 0 },
      { 0, 1, 0 },
      { 0, 0, 1 }
   };

#  if (FORMATLIST__VERSION >= 0x0104) // new examples 04/04/2003:
      std::wofstream                   wout( "woutput.txt" );

      const wchar_t                  * wstr = L"S\xF6hne Mannheim";
      std::wstring                     ws   = L"Dein Glck Liegt Mir Am Herzen";
#  endif

   // basic output examples
   // NOTE: Because 2D arrays are not supported, the 2D example does not work as intended.

   std::cout << "basic output:" << '\n' << '\n';

   std::cout << "vector< int > = "         << boost::io::formatlist( v ) << '\n';
   std::cout << "list< char > = "          << boost::io::formatlist( l ) << '\n';
   std::cout << "map< string, string > = " << boost::io::formatlist( m ) << '\n';
   std::cout << "int[ 4 ] = "              << boost::io::formatlist( i, i + 4 ) << '\n';
   std::cout << "char * = "                << boost::io::formatlist( str, str + std::strlen( str )) << '\n';
   std::cout << "std::string = "           << boost::io::formatlist( s ) << '\n';
   std::cout << "int[ 3 ][ 3 ] = "         << boost::io::formatlist( a, a + 3 ) << '\n';

#  if (FORMATLIST__VERSION >= 0x0104) // new examples 04/04/2003:
      std::cout  << "std::string = " << boost::io::formatlistex< const char * >( s ) << '\n';

      wout << L"basic output:\n\n";

      wout << L"wchar_t * = "
           << boost::io::formatlistex< wchar_t >( wstr, wstr + wcslen( wstr )) << L'\n';
      wout << L"std::wstring = "
           << boost::io::formatlistex< wchar_t >( ws ) << L'\n';
#  endif

   // list format control

   std::cout << '\n' << "formatting options:" << '\n' << '\n';

   std::cout << "vector< int > = " << boost::io::formatlist( v ).format( '(', ')' ) << '\n';
   std::cout << "vector< int > = " << boost::io::formatlist( v ).format( '|' ).space( true, true ) << '\n';

   std::cout << "map< string, string > = "
             << boost::io::formatlist( m ).format( '\n', '\0', '\n' ).space( false, false, false ) << '\n';

   std::cout << "int[ 4 ] = "
             << boost::io::formatlist( i, i + 4 ).format( ';' ).space( true, true )
             << '\n';

   std::cout << "char * = "
             << boost::io::formatlist( str, str + std::strlen( str )).format( ';' ).space( false )
             << '\n';

   std::cout << "std::string = "
             << boost::io::formatlist( s ).format( '\0', '\0', '|' ).space( false, false, false )
             << '\n';

   std::cout << '\n' << "sub-list:" << '\n' << '\n';

   std::cout << "std::string = "
             << boost::io::formatlist( s.begin() + 6, s.end()).space( false, false, false ).format( '"', '"', '-' )
             << '\n';

#  if (FORMATLIST__VERSION >= 0x0104) // new examples 04/04/2003:
      std::cout << '\n' << "advanced formatting:" << '\n' << '\n';

      std::cout << "std::string = "
                << boost::io::formatlistex< const char * >( s ).format( "<:", ":>", "\0" ) << '\n';

      std::cout << "std::string = "
                << boost::io::formatlistex< const char * >( s )
                   .format( "{ '", "' }", "', '" )
                   .space( false, false, false )
                << '\n';

      std::cout << "int[ 4 ] = "
                << boost::io::formatlistex< const char * >( i, i + 4 ).format( "-:-" ).space( true, true ) << '\n';

      std::cout << "int[ 4 ] = "
                << boost::io::formatlistex< const char * >( i, i + 4 )
                   .format( "-\n-" ).format( "\n-", "-" )
                   .space( true, true )
                << '\n';

      // This example uses char * instead of const char *:
      // there is no real distinction between them in this case, but char * is shorter to type!

      // NOTE: the same applies for wchar_t * and const wchar_t *

      std::cout << "map< string, string > = "
                << boost::io::formatlistex< char * >( m ).format( "\n==>", "\0", "\n==>" ).space( false, true, true ) << '\n';

      wout << L'\n' << L"advanced formatting:" << L'\n' << L'\n';

      wout << L"std::wstring = "
           << boost::io::formatlistex< wchar_t >( ws )
#          if !defined(EXTENDED_CHARACTER_BUG)
              .format( L'\x2202' ).space( true, false )     // partial differential
              .format( L'\x3008', L'\x3009' ).space( true ) // left and right angle brackets
#          else
              .format( L'<', L'>' ).space( true ) // left and right angle brackets (ascii versions)
#          endif
           << L'\n';

      wout << L"std::wstring = "
           << boost::io::formatlistex< wchar_t >( ws )
#          if !defined(EXTENDED_CHARACTER_BUG)
              .format( L'\x5FC3' ).space( true, true ) // chinese character: heart
#          else
              .format( L'+' ).space( true, true )
#          endif
           << L'\n';

      wout << L"std::list< std::wstring > = "
           << boost::io::formatlistex< wchar_t * >( wl )
#          if !defined(EXTENDED_CHARACTER_BUG)
              .format( L"\n  ", L"\0", L"\n\x21D2 " ).space( false, false, false ) // implies (rightwards double arrow)
#          else
              .format( L"\n    ", L"\0", L"\n==> " ).space( false, false, false )
#          endif
           << L'\n';
#  endif

   // advanced list format control

#  if 0
      /*
         The following examples are not yet supported by the format list library and thus will not
         compile. They are intended to show one of the areas where I am looking at extending the
         library so I can get preliminary feedback.

         It is important to note that anything included here is subject to change.
      */

      // possible form of a list format configuration class:

      boost::io::formatter< char >     fmt;
      fmt.format( '(', ')' ); // customize the format type

      // one possible mechanism to bind the format configuration to a formatlist_t objecxt:

      std::cout << "int[ 4 ] = " << boost::io::formatlist( i, i + 4 ).format( fmt ) << '\n';
      std::cout << "char * = "   << boost::io::formatlist( str, str + std::strlen( str )).format( fmt ) << '\n';

      // another possible mechanism:
      
      // NOTE: In this instance, the formatlist_t object would inherit fmt's FormatType value;
      // that is, if fmt uses wchar_t * values for formatting, so will the formatlist_t object.

      std::cout << "vector< int > = " << boost::io::formatlist( v, fmt ) << '\n';
#  endif

   return( 0 );
}
